import logging
import math
import wpilib

_getFPGATimestamp = wpilib.Timer.getFPGATimestamp


class LoopTimer:
    """
    A utility class that measures the number of loops that a robot program
    executes, and computes the min/max/average period for loops in the last
    second.

    Example usage::

        class Robot(wpilib.TimedRobot):

            def teleopInit(self):
                self.loop_timer = LoopTimer(self.logger)

            def teleopPeriodic(self):
                self.loop_timer.measure()

    Mainly intended for debugging purposes to measure how much lag.
    """

    def __init__(self, logger: logging.Logger):
        self.logger = logger
        self.timer = wpilib.Timer()
        self.timer.start()

        self.reset()

    def reset(self) -> None:
        self.timer.reset()

        self.start = self.last = _getFPGATimestamp()
        self.min_time = math.inf
        self.max_time = -1
        self.loops = 0

    def measure(self) -> None:
        """
        Computes loop performance information and periodically dumps it to
        the info logger.
        """

        # compute min/max/count
        now = _getFPGATimestamp()
        diff = now - self.last
        self.min_time = min(self.min_time, diff)
        self.max_time = max(self.max_time, diff)

        self.loops += 1
        self.last = now

        if self.timer.advanceIfElapsed(1):
            self.logger.info(
                "Loops: %d; min: %.3f; max: %.3f; period: %.3f; avg: %.3f",
                self.loops,
                self.min_time,
                self.max_time,
                now - self.start,
                (now - self.start) / self.loops,
            )

            self.min_time = math.inf
            self.max_time = -1
            self.start = now
            self.loops = 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os

from os.path import abspath, join, dirname

# Insert module path here
sys.path.insert(0, abspath(dirname(__file__)))
sys.path.insert(0, abspath(join(dirname(__file__), "..")))

import robotpy_ext

# -- RTD configuration ------------------------------------------------

# on_rtd is whether we are on readthedocs.org, this line of code grabbed from docs.readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

# This is used for linking and such so we link to the thing we're building
rtd_version = os.environ.get("READTHEDOCS_VERSION", "latest")
if rtd_version not in ["stable", "latest"]:
    rtd_version = "stable"

# -- General configuration ------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx_autodoc_typehints",
]

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "RobotPy WPILib Utilities"
copyright = "2015, RobotPy development team"

intersphinx_mapping = {
    "commandsv1": (
        "https://robotpy.readthedocs.io/projects/commands-v1/en/%s/" % rtd_version,
        None,
    ),
    "networktables": (
        "https://robotpy.readthedocs.io/projects/pynetworktables/en/%s/" % rtd_version,
        None,
    ),
    "wpilib": (
        "https://robotpy.readthedocs.io/projects/wpilib/en/%s/" % rtd_version,
        None,
    ),
}

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = ".".join(robotpy_ext.__version__.split(".")[:2])
# The full version, including alpha/beta/rc tags.
release = robotpy_ext.__version__

autoclass_content = "both"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output ----------------------------------------------

html_theme = "sphinx_rtd_theme"

# Output file base name for HTML help builder.
htmlhelp_basename = "sphinxdoc"

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [("index", "sphinx.tex", ". Documentation", "Author", "manual")]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [("index", "sphinx", ". Documentation", ["Author"], 1)]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "sphinx",
        ". Documentation",
        "Author",
        "sphinx",
        "One line description of project.",
        "Miscellaneous",
    )
]

# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = "."
epub_author = "Author"
epub_publisher = "Author"
epub_copyright = "2015, Author"

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]

# -- Custom Document processing ----------------------------------------------

from robotpy_sphinx.sidebar import generate_sidebar

generate_sidebar(
    globals(),
    "utilities",
    "https://raw.githubusercontent.com/robotpy/docs-sidebar/master/sidebar.toml",
)



from robotpy_ext.common_drivers import units


def close_enough(val_1, val_2):
    """Quick function to determine if val_1 ~= val_2"""
    return abs(val_1 - val_2) < 0.001


def test_meters_to_cm():
    assert units.convert(units.meter, units.centimeter, 1) == 100
    assert units.convert(units.meter, units.centimeter, 50) == 5000


def test_cm_to_meters():
    assert units.convert(units.centimeter, units.meter, 100) == 1
    assert units.convert(units.centimeter, units.meter, 500) == 5


def test_feet_to_inches():
    assert close_enough(units.convert(units.foot, units.inch, 1), 12)
    assert close_enough(units.convert(units.foot, units.inch, 4), 48)


def test_inches_to_feet():
    assert close_enough(units.convert(units.inch, units.foot, 12), 1)
    assert close_enough(units.convert(units.inch, units.foot, 48), 4)


def test_feet_to_cm():
    assert close_enough(units.convert(units.foot, units.centimeter, 1), 30.48)
    assert close_enough(units.convert(units.foot, units.centimeter, 3), 91.44)


def test_cm_to_feet():
    assert close_enough(units.convert(units.centimeter, units.foot, 30.48), 1)
    assert close_enough(units.convert(units.centimeter, units.foot, 91.44), 3)



import contextlib
import inspect
import logging
import sys
import types
import typing

from typing import Any, Callable, Dict, List, Tuple

import hal
import wpilib

from ntcore import NetworkTableInstance

# from wpilib.shuffleboard import Shuffleboard

from robotpy_ext.autonomous import AutonomousModeSelector
from robotpy_ext.misc import NotifierDelay
from robotpy_ext.misc.simple_watchdog import SimpleWatchdog

from .inject import get_injection_requests, find_injections
from .magic_tunable import setup_tunables, tunable, collect_feedbacks
from .magic_reset import collect_resets

__all__ = ["MagicRobot"]


class MagicInjectError(ValueError):
    pass


class MagicRobot(wpilib.RobotBase):
    """
    Robots that use the MagicBot framework should use this as their
    base robot class. If you use this as your base, you must
    implement the following methods:

    - :meth:`createObjects`
    - :meth:`teleopPeriodic`

    MagicRobot uses the :class:`.AutonomousModeSelector` to allow you
    to define multiple autonomous modes and to select one of them via
    the SmartDashboard/Shuffleboard.

    MagicRobot will set the following NetworkTables variables
    automatically:

    - ``/robot/mode``: one of 'disabled', 'auto', 'teleop', or 'test'
    - ``/robot/is_simulation``: True/False
    - ``/robot/is_ds_attached``: True/False

    """

    #: Amount of time each loop takes (default is 20ms)
    control_loop_wait_time = 0.020

    #: Error report interval: when an FMS is attached, how often should
    #: uncaught exceptions be reported?
    error_report_interval = 0.5

    #: A Python logging object that you can use to send messages to the log.
    #: It is recommended to use this instead of print statements.
    logger = logging.getLogger("robot")

    #: If True, teleopPeriodic will be called in autonomous mode
    use_teleop_in_autonomous = False

    def __init__(self) -> None:
        super().__init__()

        self._exclude_from_injection = ["logger"]

        self.__last_error_report = -10

        self._components: List[Tuple[str, Any]] = []
        self._feedbacks: List[Tuple[Callable[[], Any], Callable[[Any], Any]]] = []
        self._reset_components: List[Tuple[Dict[str, Any], Any]] = []

        self.__done = False

        # cache these
        self.__is_ds_attached = wpilib.DriverStation.isDSAttached
        self.__sd_update = wpilib.SmartDashboard.updateValues
        self.__lv_update = wpilib.LiveWindow.updateValues
        # self.__sf_update = Shuffleboard.update

    def _simulationInit(self) -> None:
        pass

    def _simulationPeriodic(self) -> None:
        pass

    def __simulationPeriodic(self) -> None:
        hal.simPeriodicBefore()
        self._simulationPeriodic()
        hal.simPeriodicAfter()

    def robotInit(self) -> None:
        """
        .. warning:: Internal API, don't override; use :meth:`createObjects` instead
        """

        # Create the user's objects and stuff here
        self.createObjects()

        # Load autonomous modes
        self._automodes = AutonomousModeSelector("autonomous")

        # Next, create the robot components and wire them together
        self._create_components()

        # cache these
        self.__is_ds_attached = wpilib.DriverStation.isDSAttached
        self.__sd_update = wpilib.SmartDashboard.updateValues
        self.__lv_update = wpilib.LiveWindow.updateValues
        # self.__sf_update = Shuffleboard.update

        self.__nt = NetworkTableInstance.getDefault().getTable("/robot")

        self.__nt_put_is_ds_attached = self.__nt.getEntry("is_ds_attached").setBoolean
        self.__nt_put_mode = self.__nt.getEntry("mode").setString

        self.__nt.putBoolean("is_simulation", self.isSimulation())
        self.__nt_put_is_ds_attached(self.__is_ds_attached())

        self.watchdog = SimpleWatchdog(self.control_loop_wait_time)

        self.__periodics: List[Tuple[Callable[[], None], str]] = [
            (self.robotPeriodic, "robotPeriodic()"),
        ]

        if self.isSimulation():
            self._simulationInit()
            self.__periodics.append((self.__simulationPeriodic, "simulationPeriodic()"))

    def createObjects(self) -> None:
        """
        You should override this and initialize all of your wpilib
        objects here (and not in your components, for example). This
        serves two purposes:

        - It puts all of your motor/sensor initialization in the same
          place, so that if you need to change a port/pin number it
          makes it really easy to find it. Additionally, if you want
          to create a simplified robot program to test a specific
          thing, it makes it really easy to copy/paste it elsewhere

        - It allows you to use the magic injection mechanism to share
          variables between components

        .. note:: Do not access your magic components in this function,
                  as their instances have not been created yet. Do not
                  create them either.
        """
        raise NotImplementedError

    def autonomousInit(self) -> None:
        """Initialization code for autonomous mode may go here.

        Users may override this method for initialization code which
        will be called each time the robot enters autonomous mode,
        regardless of the selected autonomous mode.

        This can be useful for code that must be run at the beginning of a match.

        .. note::

           This method is called after every component's ``on_enable`` method,
           but before the selected autonomous mode's ``on_enable`` method.
        """
        pass

    def teleopInit(self) -> None:
        """
        Initialization code for teleop control code may go here.

        Users may override this method for initialization code which will be
        called each time the robot enters teleop mode.

        .. note:: The ``on_enable`` functions of all components are called
                  before this function is called.
        """
        pass

    def teleopPeriodic(self):
        """
        Periodic code for teleop mode should go here.

        Users should override this method for code which will be called
        periodically at a regular rate while the robot is in teleop mode.

        This code executes before the ``execute`` functions of all
        components are called.

        .. note:: If you want this function to be called in autonomous
                  mode, set ``use_teleop_in_autonomous`` to True in your
                  robot class.
        """
        func = self.teleopPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.warning(
                "Default MagicRobot.teleopPeriodic() method... Override me!"
            )
            func.firstRun = False

    def disabledInit(self) -> None:
        """
        Initialization code for disabled mode may go here.

        Users may override this method for initialization code which will be
        called each time the robot enters disabled mode.

        .. note:: The ``on_disable`` functions of all components are called
                  before this function is called.
        """
        pass

    def disabledPeriodic(self):
        """
        Periodic code for disabled mode should go here.

        Users should override this method for code which will be called
        periodically at a regular rate while the robot is in disabled mode.

        This code executes before the ``execute`` functions of all
        components are called.
        """
        func = self.disabledPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.warning(
                "Default MagicRobot.disabledPeriodic() method... Override me!"
            )
            func.firstRun = False

    def testInit(self) -> None:
        """Initialization code for test mode should go here.

        Users should override this method for initialization code which will be
        called each time the robot enters disabled mode.
        """
        pass

    def testPeriodic(self) -> None:
        """Periodic code for test mode should go here."""
        pass

    def robotPeriodic(self) -> None:
        """
        Periodic code for all modes should go here.

        Users must override this method to utilize it
        but it is not required.

        This function gets called last in each mode.
        You may use it for any code you need to run
        during all modes of the robot (e.g NetworkTables updates)

        The default implementation will update
        SmartDashboard, LiveWindow and Shuffleboard.
        """
        watchdog = self.watchdog
        self.__sd_update()
        watchdog.addEpoch("SmartDashboard")
        self.__lv_update()
        watchdog.addEpoch("LiveWindow")
        # self.__sf_update()
        # watchdog.addEpoch("Shuffleboard")

    def onException(self, forceReport: bool = False) -> None:
        """
        This function must *only* be called when an unexpected exception
        has occurred that would otherwise crash the robot code. Use this
        inside your :meth:`operatorActions` function.

        If the FMS is attached (eg, during a real competition match),
        this function will return without raising an error. However,
        it will try to report one-off errors to the Driver Station so
        that it will be recorded in the Driver Station Log Viewer.
        Repeated errors may not get logged.

        Example usage::

            def teleopPeriodic(self):
                try:
                    if self.joystick.getTrigger():
                        self.shooter.shoot()
                except:
                    self.onException()

                try:
                    if self.joystick.getRawButton(2):
                        self.ball_intake.run()
                except:
                    self.onException()

                # and so on...

        :param forceReport: Always report the exception to the DS. Don't
                            set this to True
        """
        # If the FMS is not attached, crash the robot program
        if not wpilib.DriverStation.isFMSAttached():
            raise

        # Otherwise, if the FMS is attached then try to report the error via
        # the driver station console. Maybe.
        now = wpilib.Timer.getFPGATimestamp()

        try:
            if (
                forceReport
                or (now - self.__last_error_report) > self.error_report_interval
            ):
                wpilib.reportError("Unexpected exception", True)
        except:
            pass  # ok, can't do anything here

        self.__last_error_report = now

    @contextlib.contextmanager
    def consumeExceptions(self, forceReport: bool = False):
        """
        This returns a context manager which will consume any uncaught
        exceptions that might otherwise crash the robot.

        Example usage::

            def teleopPeriodic(self):
                with self.consumeExceptions():
                    if self.joystick.getTrigger():
                        self.shooter.shoot()

                with self.consumeExceptions():
                    if self.joystick.getRawButton(2):
                        self.ball_intake.run()

                # and so on...

        :param forceReport: Always report the exception to the DS. Don't
                            set this to True

        .. seealso:: :meth:`onException` for more details
        """
        try:
            yield
        except:
            self.onException(forceReport=forceReport)

    #
    # Internal API
    #

    def startCompetition(self) -> None:
        """
        This runs the mode-switching loop.

        .. warning:: Internal API, don't override
        """

        # TODO: usage reporting?
        self.robotInit()

        # Tell the DS the robot is ready to be enabled
        hal.observeUserProgramStarting()

        while not self.__done:
            isEnabled, isAutonomous, isTest = self.getControlState()

            if not isEnabled:
                self._disabled()
            elif isAutonomous:
                self.autonomous()
            elif isTest:
                self._test()
            else:
                self._operatorControl()

    def endCompetition(self) -> None:
        self.__done = True
        self._automodes.endCompetition()

    def autonomous(self) -> None:
        """
        MagicRobot will do The Right Thing and automatically load all
        autonomous mode routines defined in the autonomous folder.

        .. warning:: Internal API, don't override
        """

        self.__nt_put_mode("auto")
        self.__nt_put_is_ds_attached(self.__is_ds_attached())

        self._on_mode_enable_components()

        try:
            self.autonomousInit()
        except:
            self.onException(forceReport=True)

        auto_functions: Tuple[Callable[[], None], ...] = (self._enabled_periodic,)

        if self.use_teleop_in_autonomous:
            auto_functions = (self.teleopPeriodic,) + auto_functions

        self._automodes.run(
            self.control_loop_wait_time,
            auto_functions,
            self.onException,
            watchdog=self.watchdog,
        )

        self._on_mode_disable_components()

    def _disabled(self) -> None:
        """
        This function is called in disabled mode. You should not
        override this function; rather, you should override the
        :meth:`disabledPeriodic` function instead.

        .. warning:: Internal API, don't override
        """
        watchdog = self.watchdog
        watchdog.reset()

        self.__nt_put_mode("disabled")
        ds_attached = None

        self._on_mode_disable_components()
        try:
            self.disabledInit()
        except:
            self.onException(forceReport=True)
        watchdog.addEpoch("disabledInit()")

        refreshData = wpilib.DriverStation.refreshData
        DSControlWord = wpilib.DSControlWord

        with NotifierDelay(self.control_loop_wait_time) as delay:
            while not self.__done:
                refreshData()
                cw = DSControlWord()
                if cw.isEnabled():
                    break

                if ds_attached != cw.isDSAttached():
                    ds_attached = not ds_attached
                    self.__nt_put_is_ds_attached(ds_attached)

                hal.observeUserProgramDisabled()
                try:
                    self.disabledPeriodic()
                except:
                    self.onException()
                watchdog.addEpoch("disabledPeriodic()")

                self._do_periodics()
                # watchdog.disable()
                watchdog.printIfExpired()

                delay.wait()
                watchdog.reset()

    def _operatorControl(self) -> None:
        """
        This function is called in teleoperated mode. You should not
        override this function; rather, you should override the
        :meth:`teleopPeriodics` function instead.

        .. warning:: Internal API, don't override
        """
        watchdog = self.watchdog
        watchdog.reset()

        self.__nt_put_mode("teleop")
        # don't need to update this during teleop -- presumably will switch
        # modes when ds is no longer attached
        self.__nt_put_is_ds_attached(self.__is_ds_attached())

        # initialize things
        self._on_mode_enable_components()

        try:
            self.teleopInit()
        except:
            self.onException(forceReport=True)
        watchdog.addEpoch("teleopInit()")

        observe = hal.observeUserProgramTeleop
        refreshData = wpilib.DriverStation.refreshData
        isTeleopEnabled = wpilib.DriverStation.isTeleopEnabled

        with NotifierDelay(self.control_loop_wait_time) as delay:
            while not self.__done:
                refreshData()
                if not isTeleopEnabled():
                    break

                observe()
                try:
                    self.teleopPeriodic()
                except:
                    self.onException()
                watchdog.addEpoch("teleopPeriodic()")

                self._enabled_periodic()
                # watchdog.disable()
                watchdog.printIfExpired()

                delay.wait()
                watchdog.reset()

        self._on_mode_disable_components()

    def _test(self) -> None:
        """Called when the robot is in test mode"""
        watchdog = self.watchdog
        watchdog.reset()

        self.__nt_put_mode("test")
        self.__nt_put_is_ds_attached(self.__is_ds_attached())

        wpilib.LiveWindow.setEnabled(True)
        # Shuffleboard.enableActuatorWidgets()

        try:
            self.testInit()
        except:
            self.onException(forceReport=True)
        watchdog.addEpoch("testInit()")

        refreshData = wpilib.DriverStation.refreshData
        DSControlWord = wpilib.DSControlWord

        with NotifierDelay(self.control_loop_wait_time) as delay:
            while not self.__done:
                refreshData()
                cw = DSControlWord()
                if not (cw.isTest() and cw.isEnabled()):
                    break

                hal.observeUserProgramTest()
                try:
                    self.testPeriodic()
                except:
                    self.onException()
                watchdog.addEpoch("testPeriodic()")

                self._do_periodics()
                # watchdog.disable()
                watchdog.printIfExpired()

                delay.wait()
                watchdog.reset()

        wpilib.LiveWindow.setEnabled(False)
        # Shuffleboard.disableActuatorWidgets()

    def _on_mode_enable_components(self) -> None:
        # initialize things
        for _, component in self._components:
            on_enable = getattr(component, "on_enable", None)
            if on_enable is not None:
                try:
                    on_enable()
                except:
                    self.onException(forceReport=True)

    def _on_mode_disable_components(self) -> None:
        # deinitialize things
        for _, component in self._components:
            on_disable = getattr(component, "on_disable", None)
            if on_disable is not None:
                try:
                    on_disable()
                except:
                    self.onException(forceReport=True)

    def _create_components(self) -> None:
        #
        # TODO: Will need to inject into any autonomous mode component
        #       too, as they're a bit different
        #

        # TODO: Will need to order state machine components before
        #       other components just in case

        components = []

        self.logger.info("Creating magic components")

        # Identify all of the types, and create them
        cls = type(self)

        # - Iterate over class variables with type annotations
        # .. this hack is necessary for pybind11 based modules
        sys.modules["pybind11_builtins"] = types.SimpleNamespace()  # type: ignore

        injectables = self._collect_injectables()

        for m, ctyp in typing.get_type_hints(cls).items():
            # Ignore private variables
            if m.startswith("_"):
                continue

            # If the variable has been set, skip it
            if hasattr(self, m):
                continue

            # If the type is not actually a type, give a meaningful error
            if not isinstance(ctyp, type):
                raise TypeError(
                    "%s has a non-type annotation on %s (%r); lone non-injection variable annotations are disallowed, did you want to assign a static variable?"
                    % (cls.__name__, m, ctyp)
                )

            component = self._create_component(m, ctyp, injectables)

            # Store for later
            components.append((m, component))
            injectables[m] = component

        # For each new component, perform magic injection
        for cname, component in components:
            setup_tunables(component, cname, "components")
            self._setup_vars(cname, component, injectables)
            self._setup_reset_vars(component)

        # Do it for autonomous modes too
        for mode in self._automodes.modes.values():
            mode.logger = logging.getLogger(mode.MODE_NAME)
            setup_tunables(mode, mode.MODE_NAME, "autonomous")
            self._setup_vars(mode.MODE_NAME, mode, injectables)

        # And for self too
        setup_tunables(self, "robot", None)
        self._feedbacks += collect_feedbacks(self, "robot", None)

        # Call setup functions for components
        for cname, component in components:
            setup = getattr(component, "setup", None)
            if setup is not None:
                setup()
            # ... and grab all the feedback methods
            self._feedbacks += collect_feedbacks(component, cname, "components")

        # Call setup functions for autonomous modes
        for mode in self._automodes.modes.values():
            if hasattr(mode, "setup"):
                mode.setup()

        self._components = components

    def _collect_injectables(self) -> Dict[str, Any]:
        injectables = {}
        cls = type(self)

        for n in dir(self):
            if (
                n.startswith("_")
                or n in self._exclude_from_injection
                or isinstance(getattr(cls, n, None), (property, tunable))
            ):
                continue

            o = getattr(self, n)

            # Don't inject methods
            # TODO: This could actually be a cool capability..
            if inspect.ismethod(o):
                continue

            injectables[n] = o

        return injectables

    def _create_component(self, name: str, ctyp: type, injectables: Dict[str, Any]):
        type_hints = typing.get_type_hints(ctyp.__init__)
        NoneType = type(None)
        init_return_type = type_hints.pop("return", NoneType)
        assert (
            init_return_type is NoneType
        ), f"{ctyp!r} __init__ had an unexpected non-None return type hint"
        requests = get_injection_requests(type_hints, name)
        injections = find_injections(requests, injectables, name)

        # Create instance, set it on self
        component = ctyp(**injections)
        setattr(self, name, component)

        # Ensure that mandatory methods are there
        if not callable(getattr(component, "execute", None)):
            raise ValueError(
                "Component %s (%r) must have a method named 'execute'"
                % (name, component)
            )

        # Automatically inject a logger object
        component.logger = logging.getLogger(name)

        self.logger.info("-> %s (class: %s)", name, ctyp.__name__)

        return component

    def _setup_vars(self, cname: str, component, injectables: Dict[str, Any]) -> None:
        self.logger.debug("Injecting magic variables into %s", cname)

        type_hints = typing.get_type_hints(type(component))
        requests = get_injection_requests(type_hints, cname, component)
        injections = find_injections(requests, injectables, cname)
        component.__dict__.update(injections)

    def _setup_reset_vars(self, component) -> None:
        reset_dict = collect_resets(type(component))

        if reset_dict:
            component.__dict__.update(reset_dict)
            self._reset_components.append((reset_dict, component))

    def _do_periodics(self) -> None:
        """Run periodic methods which run in every mode."""
        watchdog = self.watchdog

        for method, setter in self._feedbacks:
            try:
                value = method()
            except:
                self.onException()
            else:
                setter(value)

        watchdog.addEpoch("@magicbot.feedback")

        for periodic, name in self.__periodics:
            periodic()
            watchdog.addEpoch(name)

    def _enabled_periodic(self) -> None:
        """Run components and all periodic methods."""
        watchdog = self.watchdog

        for name, component in self._components:
            try:
                component.execute()
            except:
                self.onException()
            watchdog.addEpoch(name)

        self._do_periodics()

        for reset_dict, component in self._reset_components:
            component.__dict__.update(reset_dict)



import wpilib
import math


class SharpIR2Y0A02:
    """
    Sharp GP2Y0A02YK0F is an analog IR sensor capable of measuring
    distances from 20cm to 150cm. Output distance is measured in
    centimeters.

    Distance is calculated using the following equation derived from
    the graph provided in the datasheet::

        62.28*x ^ -1.092

    .. warning:: FRC Teams: the case on these sensors is conductive and
                 grounded, and should not be mounted on a metallic
                 surface!
    """

    def __init__(self, port):
        """:param port: Analog port number"""
        self.distance = wpilib.AnalogInput(port)

    def getDistance(self):
        """
        :returns: distance in centimeters. The output is constrained to
                  be between 22.5 and 145
        """

        # Don't allow zero/negative values
        v = max(self.distance.getVoltage(), 0.00001)
        d = 62.28 * math.pow(v, -1.092)

        # Constrain output
        return max(min(d, 145.0), 22.5)


class SharpIR2Y0A21:
    """
    Sharp GP2Y0A21YK0F is an analog IR sensor capable of measuring
    distances from 10cm to 80cm. Output distance is measured in
    centimeters.

    Distance is calculated using the following equation derived from
    the graph provided in the datasheet::

        26.449*x ^ -1.226

    .. warning:: FRC Teams: the case on these sensors is conductive and
                 grounded, and should not be mounted on a metallic
                 surface!
    """

    def __init__(self, port):
        """:param port: Analog port number"""
        self.distance = wpilib.AnalogInput(port)

    def getDistance(self):
        """
        :returns: distance in centimeters. The output is constrained to
                  be between 10 and 80
        """

        # Don't allow zero/negative values
        v = max(self.distance.getVoltage(), 0.00001)
        d = 26.449 * math.pow(v, -1.226)

        # Constrain output
        return max(min(d, 80.0), 10.0)


class SharpIR2Y0A41:
    """
    Sharp GP2Y0A41SK0F is an analog IR sensor capable of measuring
    distances from 4cm to 40cm. Output distance is measured in
    centimeters.

    Distance is calculated using the following equation derived from
    the graph provided in the datasheet::

        12.84*x ^ -0.9824

    .. warning:: FRC Teams: the case on these sensors is conductive and
                 grounded, and should not be mounted on a metallic
                 surface!
    """

    def __init__(self, port):
        """:param port: Analog port number"""
        self.distance = wpilib.AnalogInput(port)

    def getDistance(self):
        """
        :returns: distance in centimeters. The output is constrained to
                  be between 4.5 and 35
        """

        # Don't allow zero/negative values
        v = max(self.distance.getVoltage(), 0.00001)
        d = 12.84 * math.pow(v, -0.9824)

        # Constrain output
        return max(min(d, 35.0), 4.5)


# backwards compat
SharpIRGP2Y0A41SK0F = SharpIR2Y0A41



from typing import Any, Generic, TypeVar, overload

_V = TypeVar("_V")

class will_reset_to(Generic[_V]):
    default: _V

    def __init__(self, default: _V) -> None: ...
    # we don't really have __get__, but this makes code in methods using these
    # to type-check, whilst giving correct behaviour for code at the class level
    @overload
    def __get__(self, instance: None, owner=...) -> will_reset_to: ...
    @overload
    def __get__(self, instance, owner=...) -> _V: ...
    def __set__(self, instance, value: _V) -> None: ...

def collect_resets(cls: type) -> dict[str, Any]: ...



from .magicrobot import MagicRobot
from .magic_tunable import feedback, tunable
from .magic_reset import will_reset_to

from .state_machine import (
    AutonomousStateMachine,
    StateMachine,
    default_state,
    state,
    timed_state,
)

__all__ = (
    "MagicRobot",
    "feedback",
    "tunable",
    "will_reset_to",
    "AutonomousStateMachine",
    "StateMachine",
    "default_state",
    "state",
    "timed_state",
)



import importlib
import inspect
import logging
import os
from glob import glob
from typing import Callable, Sequence, Union

import hal
import wpilib

from ..misc.precise_delay import NotifierDelay
from ..misc.simple_watchdog import SimpleWatchdog

logger = logging.getLogger("autonomous")


class AutonomousModeSelector:
    """
    This object loads all modules in a specified python package, and tries
    to automatically discover autonomous modes from them. Each module is
    added to a ``SendableChooser`` object, which allows the user to select
    one of them via SmartDashboard.

    Autonomous mode objects must implement the following functions:

    - ``on_enable`` - Called when autonomous mode is initially enabled
    - ``on_disable`` - Called when autonomous mode is no longer active
    - ``on_iteration`` - Called for each iteration of the autonomous control loop

    Your autonomous object may have the following attributes:

    - ``MODE_NAME`` - The name of the autonomous mode to display to users
    - ``DISABLED`` - If True, don't allow this mode to be selected
    - ``DEFAULT`` - If True, this is the default autonomous mode selected

    Here is an example of using ``AutonomousModeSelector`` in ``TimedRobot``:

    .. code-block:: python

        class MyRobot(wpilib.TimedRobot):

            def robotInit(self):
                self.automodes = AutonomousModeSelector('autonomous')

            def autonomousInit(self):
                self.automodes.start()

            def autonomousPeriodic(self):
                self.automodes.periodic()

            def disabledInit(self):
                self.automodes.disable()

    If you use AutonomousModeSelector, you may also be interested in
    the autonomous state machine helper (:class:`.StatefulAutonomous`).

    Check out the samples in our github repository that show some basic
    usage of ``AutonomousModeSelector``.

    .. note:: If you use AutonomousModeSelector, then you should add
              ``robotpy_ext.autonomous.selector_tests`` to your pyfrc
              unit tests like so::

                  from robotpy_ext.autonomous.selector_tests import *

    .. note::

       For your autonomous mode's ``on_disable`` method to be called,
       you must call :meth:`disable` in ``disabledInit``.

       It is okay to not call :meth:`disable` if you do not need ``on_disable``.
    """

    def __init__(self, autonomous_pkgname, *args, **kwargs):
        """
        :param autonomous_pkgname: Module to load autonomous modes from
        :param args: Args to pass to created autonomous modes
        :param kwargs: Keyword args to pass to created autonomous modes
        """

        self.modes = {}
        self.active_mode = None
        self.robot_exit = False

        logger.info("Begin initializing autonomous mode switcher")

        # load all modules in specified module
        modules = []

        try:
            autonomous_pkg = importlib.import_module(autonomous_pkgname)
        except ImportError as e:
            if e.name not in [autonomous_pkgname, autonomous_pkgname.split(".")[0]]:
                raise

            # Don't kill the robot because they didn't create an autonomous package
            logger.warning("Cannot load the '%s' package", autonomous_pkgname)
        else:
            pkgdirs = []
            pkgfile = getattr(autonomous_pkg, "__file__", None)
            if pkgfile:
                pkgdirs = [os.path.dirname(os.path.abspath(pkgfile))]
            else:
                # implicit packages have no __file__, just __path__
                pkgpath = getattr(autonomous_pkg, "__path__", None)
                if pkgpath:
                    pkgdirs = list(set(pkgpath))

            if pkgdirs:
                for pkgdir in pkgdirs:
                    modules.extend(glob(os.path.join(pkgdir, "*.py")))

        for module_filename in modules:
            module = None
            module_name = os.path.basename(module_filename[:-3])

            if module_name in ["__init__"]:
                continue

            try:
                module = importlib.import_module("." + module_name, autonomous_pkgname)
                # module = imp.load_source('.' + module_name, module_filename)
            except:
                if not wpilib.DriverStation.isFMSAttached():
                    raise

            #
            # Find autonomous mode classes in the modules that are present
            # -> note that we actually create the instance of the objects here,
            #    so that way we find out about any errors *before* we get out
            #    on the field..

            for name, obj in inspect.getmembers(module, inspect.isclass):
                mode_name = getattr(obj, "MODE_NAME", None)
                if mode_name is not None:
                    # don't allow the driver to select this mode
                    if getattr(obj, "DISABLED", False):
                        logger.warning(
                            "autonomous mode %s is marked as disabled", obj.MODE_NAME
                        )
                        continue

                    try:
                        instance = obj(*args, **kwargs)
                    except:
                        if not wpilib.DriverStation.isFMSAttached():
                            raise
                        else:
                            continue

                    if mode_name in self.modes:
                        if not wpilib.DriverStation.isFMSAttached():
                            raise RuntimeError(
                                "Duplicate name %s in %s" % (mode_name, module_filename)
                            )

                        logger.error(
                            "Duplicate name %s specified by object type %s in module %s",
                            mode_name,
                            name,
                            module_filename,
                        )
                        self.modes[name + "_" + module_filename] = instance
                    else:
                        self.modes[instance.MODE_NAME] = instance

        # now that we have a bunch of valid autonomous mode objects, let
        # the user select one using the SmartDashboard.

        # SmartDashboard interface
        self.chooser = wpilib.SendableChooser()

        default_modes = []
        mode_names = []

        logger.info("Loaded autonomous modes:")
        for k, v in sorted(self.modes.items()):
            if getattr(v, "DEFAULT", False):
                logger.info(" -> %s [Default]", k)
                self.chooser.setDefaultOption(k, v)
                default_modes.append(k)
            else:
                logger.info(" -> %s", k)
                self.chooser.addOption(k, v)

            mode_names.append(k)

        if len(self.modes) == 0:
            logger.warning("-- no autonomous modes were loaded!")

        self.chooser.addOption("None", None)

        if len(default_modes) == 0:
            self.chooser.setDefaultOption("None", None)
        elif len(default_modes) != 1:
            if not wpilib.DriverStation.isFMSAttached():
                raise RuntimeError(
                    "More than one autonomous mode was specified as default! (modes: %s)"
                    % (", ".join(default_modes))
                )

        # must PutData after setting up objects
        wpilib.SmartDashboard.putData("Autonomous Mode", self.chooser)

        # XXX: Compatibility with the FRC dashboard
        wpilib.SmartDashboard.putStringArray("Auto List", mode_names)

        logger.info("Autonomous switcher initialized")

    def endCompetition(self):
        """Call this function when your robot's endCompetition function is called"""
        self.robot_exit = True

    def run(
        self,
        control_loop_wait_time: float = 0.020,
        iter_fn: Union[Callable[[], None], Sequence[Callable[[], None]]] = None,
        on_exception: Callable = None,
        watchdog: Union[wpilib.Watchdog, SimpleWatchdog] = None,
    ) -> None:
        """
        This method implements the entire autonomous loop.

        Do not call this from ``TimedRobot`` as this will break the
        timing of your control loop when your robot switches to teleop.

        This function will NOT exit until autonomous mode has ended. If
        you need to execute code in all autonomous modes, pass a function
        or list of functions as the ``iter_fn`` parameter, and they will be
        called once per autonomous mode iteration.

        :param control_loop_wait_time: Amount of time between iterations in seconds
        :param iter_fn: Called at the end of every iteration while
                        autonomous mode is executing
        :param on_exception: Called when an uncaught exception is raised,
                             must take a single keyword arg "forceReport"
        :param watchdog: a WPILib Watchdog to feed every iteration
        """
        if watchdog is not None:
            watchdog.reset()

            if isinstance(watchdog, SimpleWatchdog):
                watchdog_check_expired = watchdog.printIfExpired
            else:

                def watchdog_check_expired():
                    if watchdog.isExpired():
                        watchdog.printEpochs()

        logger.info("Begin autonomous")

        if iter_fn is None:
            iter_fn = (lambda: None,)
        elif callable(iter_fn):
            iter_fn = (iter_fn,)

        if on_exception is None:
            on_exception = self._on_exception

        # keep track of how much time has passed in autonomous mode
        timer = wpilib.Timer()
        timer.start()

        try:
            self._on_autonomous_enable()
        except:
            on_exception(forceReport=True)
        if watchdog is not None:
            watchdog.addEpoch("auto on_enable")

        #
        # Autonomous control loop
        #

        observe = hal.observeUserProgramAutonomous
        refreshData = wpilib.DriverStation.refreshData
        isAutonomousEnabled = wpilib.DriverStation.isAutonomousEnabled

        with NotifierDelay(control_loop_wait_time) as delay:
            while not self.robot_exit:
                refreshData()
                if not isAutonomousEnabled():
                    break

                observe()
                try:
                    self._on_iteration(timer.get())
                except:
                    on_exception()
                if watchdog is not None:
                    watchdog.addEpoch("auto on_iteration")

                for fn in iter_fn:
                    fn()

                if watchdog is not None:
                    watchdog.disable()

                    watchdog_check_expired()

                delay.wait()
                if watchdog is not None:
                    watchdog.reset()

        #
        # Done with autonomous, finish up
        #

        try:
            self.disable()
        except:
            on_exception(forceReport=True)

        logger.info("Autonomous mode ended")

    def start(self) -> None:
        """Start autonomous mode.

        This initialises the selected autonomous mode.
        Call this from your ``autonomousInit`` method.

        .. versionadded:: 2020.1.5
        """
        self.timer = wpilib.Timer()
        self.timer.start()

        self._on_autonomous_enable()

    def periodic(self) -> None:
        """Execute one control loop iteration of the active autonomous mode.

        Call this from your ``autonomousPeriodic`` method.

        .. versionadded:: 2020.1.5
        """
        self._on_iteration(self.timer.get())

    def disable(self) -> None:
        """Disables the active autonomous mode.

        You can call this from your ``disabledInit`` method
        to call your autonomous mode's ``on_disable`` method.

        .. versionadded:: 2020.1.5
        """
        if self.active_mode is not None:
            logger.info("Disabling '%s'", self.active_mode.MODE_NAME)
            self.active_mode.on_disable()

        self.active_mode = None

    #
    #   Internal methods used to implement autonomous mode switching, and
    #   are called automatically
    #

    def _on_autonomous_enable(self) -> None:
        """Selects the active autonomous mode and enables it"""

        # XXX: FRC Dashboard compatibility
        # -> if you set it here, you're stuck using it. The FRC Dashboard
        #    doesn't seem to have a default (nor will it show a default),
        #    so the key will only get set if you set it.
        auto_mode = wpilib.SmartDashboard.getString("Auto Selector", None)
        if auto_mode is not None and auto_mode in self.modes:
            logger.info("Using autonomous mode set by LabVIEW dashboard")
            self.active_mode = self.modes[auto_mode]
        else:
            self.active_mode = self.chooser.getSelected()

        if self.active_mode is not None:
            logger.info("Enabling '%s'", self.active_mode.MODE_NAME)
            self.active_mode.on_enable()
        else:
            logger.warning(
                "No autonomous modes were selected, not enabling autonomous mode"
            )

    def _on_iteration(self, time_elapsed: float) -> None:
        """Run the code for the current autonomous mode"""
        if self.active_mode is not None:
            self.active_mode.on_iteration(time_elapsed)

    def _on_exception(self, forceReport: bool = False):
        if not wpilib.DriverStation.isFMSAttached():
            raise



from wpilib import AnalogInput


class REVAnalogPressureSensor:
    """
    The REV Robotics Analog Pressure Sensor is a 5V sensor
    that can measure pressures up to 200 PSI. It outputs an
    analog voltage that is proportional to the measured pressure.

    Pressure is derived from the following equation, taken from
    the data sheet::

        Pressure = 250 * (Vout / Vcc) -25

    where Vout is the output voltage of the sensor, and Vcc is the input
    voltage

    To calibrate the sensor, supply the sensor with a known
    pressure and call the calibrate method with the given
    pressure as the parameter
    """

    def __init__(self, channel, voltage_in=5):
        """
        :param voltage_in: Supply voltage to sensor from roboRIO
        """

        self.sensor = AnalogInput(channel)
        self.voltage_in = voltage_in

    @property
    def pressure(self):
        try:
            v = max(self.sensor.getAverageVoltage(), 0.00001)
            return (250 * (v / getattr(self, "Vn", self.voltage_in))) - 25
        except ZeroDivisionError:
            return 0

    def calibrate(self, known_pressure):
        """
        Calibrate calculated pressure to known pressure

        :param known_pressure: Known current pressure to calibrate to
        """

        Vo = max(self.sensor.getAverageVoltage(), 0.00001)
        self.Vn = Vo / (0.004 * known_pressure + 0.1)



import logging


class MagicComponent:
    """
    To automagically retrieve variables defined in your base robot
    object, you can add the following::

        class MyComponent:

            # other variables 'imported' automatically from MagicRobot
            elevator_motor: Talon
            other_component: MyOtherComponent

            ...

            def execute(self):

                # This will be automatically set to the Talon
                # instance created in robot.py
                self.elevator_motor.set(self.value)


    What this says is "find the variable in the robot class called
    'elevator_motor', which is a Talon". If the name and type match,
    then the variable will automatically be injected into your
    component when it is created.

    .. note:: You don't need to inherit from ``MagicComponent``, it is only
              provided for documentation's sake
    """

    logger: logging.Logger

    def setup(self) -> None:
        """
        This function is called after ``createObjects`` has been called in
        the main robot class, and after all components have been created

        The setup function is optional and components do not have to define
        one. ``setup()`` functions are called in order of component definition
        in the main robot class.

        .. note:: For technical reasons, variables imported from
                  MagicRobot are not initialized when your component's
                  constructor is called. However, they will be initialized
                  by the time this function is called.
        """

    def on_enable(self) -> None:
        """
        Called when the robot enters autonomous or teleoperated mode. This
        function should initialize your component to a "safe" state so
        that unexpected things don't happen when enabling the robot.

        .. note:: You'll note that there isn't a separate initialization
                  function for autonomous and teleoperated modes. This is
                  intentional, as they should be the same.
        """

    def on_disable(self) -> None:
        """
        Called when the robot leaves autonomous or teleoperated
        """

    def execute(self) -> None:
        """
        This function is called at the end of the control loop
        """



class DriverBase(object):
    """
    This should be the base class for all drivers in the cdl,
    currently all it does is spit out a warning message if the driver has not been verified.
    """

    #:This should be overloaded by the driver,
    # It is just a mechanism to ensure code quality for device drivers. Upon creation of a new driver,
    # it will be either left alone, or overloaded to be False. Once the functionality of the driver is verified,
    # this will get overloaded to true by whoever verifies it.
    verified = False

    def __init__(self):
        """
        Constructor for DriverBase, all this does is print a message to console if the driver has not been verified yet.
        """
        if not self.verified:
            print(
                "Warning, device driver {} has not been verified yet, please use with caution!".format(
                    self.__class__.__name__
                )
            )



import functools
import inspect
import logging
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    NoReturn,
    Optional,
    Sequence,
    Union,
    overload,
)

import wpilib

from .magic_tunable import tunable

if wpilib.RobotBase.isSimulation():
    getTime = wpilib.Timer.getFPGATimestamp
else:
    from time import monotonic as getTime


class IllegalCallError(TypeError):
    pass


class NoFirstStateError(ValueError):
    pass


class MultipleFirstStatesError(ValueError):
    pass


class MultipleDefaultStatesError(ValueError):
    pass


class InvalidStateName(ValueError):
    pass


class _State:
    def __init__(
        self,
        f: "StateMethod",
        first: bool = False,
        must_finish: bool = False,
        *,
        duration: Optional[float] = None,
        is_default: bool = False,
    ) -> None:
        name = f.__name__

        # Can't define states that are named the same as things in the
        # base class, will cause issues. Catch it early.
        if hasattr(StateMachine, name):
            raise InvalidStateName(f"cannot have a state named '{name}'")

        # inspect the args, provide a correct call implementation
        allowed_args = "self", "tm", "state_tm", "initial_call"
        sig = inspect.signature(f)
        args = []
        invalid_args = []
        for i, arg in enumerate(sig.parameters.values()):
            if i == 0 and arg.name != "self":
                raise ValueError(f"First argument to {name} must be 'self'")
            if arg.kind is arg.VAR_POSITIONAL:
                raise ValueError(f"Cannot use *args in signature for function {name}")
            if arg.kind is arg.VAR_KEYWORD:
                raise ValueError(
                    f"Cannot use **kwargs in signature for function {name}"
                )
            if arg.kind is arg.KEYWORD_ONLY:
                raise ValueError(
                    f"Cannot use keyword-only parameters for function {name}"
                )
            if arg.name in allowed_args:
                args.append(arg.name)
            else:
                invalid_args.append(arg.name)

        if invalid_args:
            raise ValueError(
                "Invalid parameter names in %s: %s" % (name, ",".join(invalid_args))
            )

        self.name = name
        self.description = inspect.getdoc(f)
        self.first = first
        self.must_finish = must_finish
        self.is_default = is_default
        self.duration = duration

        varlist = {"f": f}
        args_code = ",".join(args)
        wrapper_code = f"lambda self, tm, state_tm, initial_call: f({args_code})"
        self.run = eval(wrapper_code, varlist, varlist)

        self.next_state: Optional[StateRef]

    def __call__(self, *args, **kwargs) -> NoReturn:
        raise IllegalCallError(
            "Do not call states directly, use begin/next_state instead"
        )

    def __set_name__(self, owner: type, name: str) -> None:
        # Don't allow aliasing of states, it probably won't do what users expect
        if name != self.name:
            raise InvalidStateName(
                f"magicbot state '{self.name}' defined as attribute '{name}'"
            )

        if not issubclass(owner, StateMachine):
            raise TypeError(f"magicbot state {name} defined in non-StateMachine")

        # make durations tunable
        if self.duration is not None:
            duration_attr = name + "_duration"
            # don't create it twice (in case of inheritance overriding)
            if getattr(owner, duration_attr, None) is None:
                setattr(
                    owner,
                    duration_attr,
                    tunable(self.duration, writeDefault=False, subtable="state"),
                )


StateRef = Union[str, _State]
StateMethod = Callable[..., None]


class _StateData:
    def __init__(self, wrapper: _State) -> None:
        self.name = wrapper.name
        self.duration_attr = "%s_duration" % self.name
        self.expires: float = 0xFFFFFFFF
        self.ran = False
        self.run = wrapper.run
        self.must_finish = wrapper.must_finish

        if hasattr(wrapper, "next_state"):
            self.next_state = wrapper.next_state

        self.start_time: float


def timed_state(
    *,
    duration: float,
    next_state: Optional[StateRef] = None,
    first: bool = False,
    must_finish: bool = False,
) -> Callable[[StateMethod], _State]:
    """
    If this decorator is applied to a function in an object that inherits
    from :class:`.StateMachine`, it indicates that the function
    is a state that will run for a set amount of time unless interrupted.

    It is guaranteed that a timed_state will execute at least once, even if
    it expires prior to being executed.

    The decorated function can have the following arguments in any order:

    - ``tm`` - The number of seconds since the state machine has started
    - ``state_tm`` - The number of seconds since this state has been active
      (note: it may not start at zero!)
    - ``initial_call`` - Set to True when the state is initially called,
      False otherwise. If the state is switched to multiple times, this
      will be set to True at the start of each state execution.

    :param duration: The length of time to run the state before progressing
                     to the next state
    :param next_state: The name of the next state. If not specified, then
                       this will be the last state executed if time expires
    :param first: If True, this state will be ran first
    :param must_finish: If True, then this state will continue executing
                        even if ``engage()`` is not called. However,
                        if ``done()`` is called, execution will stop
                        regardless of whether this is set.
    """

    def decorator(f: StateMethod) -> _State:
        wrapper = _State(f, first, must_finish, duration=duration)

        wrapper.next_state = next_state

        return wrapper

    return decorator


@overload
def state(
    *,
    first: bool = ...,
    must_finish: bool = ...,
) -> Callable[[StateMethod], _State]: ...


@overload
def state(f: StateMethod) -> _State: ...


def state(
    f: Optional[StateMethod] = None,
    *,
    first: bool = False,
    must_finish: bool = False,
) -> Union[Callable[[StateMethod], _State], _State]:
    """
    If this decorator is applied to a function in an object that inherits
    from :class:`.StateMachine`, it indicates that the function
    is a state. The state will continue to be executed until the
    ``next_state`` function is executed.

    The decorated function can have the following arguments in any order:

    - ``tm`` - The number of seconds since the state machine has started
    - ``state_tm`` - The number of seconds since this state has been active
      (note: it may not start at zero!)
    - ``initial_call`` - Set to True when the state is initially called,
      False otherwise. If the state is switched to multiple times, this
      will be set to True at the start of each state execution.

    :param first: If True, this state will be ran first
    :param must_finish: If True, then this state will continue executing
                        even if ``engage()`` is not called. However,
                        if ``done()`` is called, execution will stop
                        regardless of whether this is set.
    """

    if f is None:
        return lambda f: _State(f, first, must_finish)

    return _State(f, first, must_finish)


def default_state(f: StateMethod) -> _State:
    """
    If this decorator is applied to a method in an object that inherits
    from :class:`.StateMachine`, it indicates that the method
    is a default state; that is, if no other states are executing, this
    state will execute. If the state machine is always executing, the
    default state will never execute.

    There can only be a single default state in a StateMachine object.

    The decorated function can have the following arguments in any order:

    - ``tm`` - The number of seconds since the state machine has started
    - ``state_tm`` - The number of seconds since this state has been active
      (note: it may not start at zero!)
    - ``initial_call`` - Set to True when the state is initially called,
      False otherwise. If the state is switched to multiple times, this
      will be set to True at the start of each state execution.
    """
    return _State(f, first=False, must_finish=True, is_default=True)


def _get_class_members(cls: type) -> Dict[str, Any]:
    """Get the members of the given class in definition order, bases first."""
    d = {}
    for cls in reversed(cls.__mro__):
        d.update(cls.__dict__)
    return d


class StateMachine:
    '''
    The StateMachine class is used to implement magicbot components that
    allow one to easily define a `finite state machine (FSM)
    <https://en.wikipedia.org/wiki/Finite-state_machine>`_ that can be
    executed via the magicbot framework.

    You create a component class that inherits from ``StateMachine``.
    Each state is represented as a single function, and you indicate that
    a function is a particular state by decorating it with one of the
    following decorators:

    * :func:`@default_state <.default_state>`
    * :func:`@state <.state>`
    * :func:`@timed_state <.timed_state>`

    As the state machine executes, the decorated function representing the
    current state will be called. Decorated state functions can receive the
    following parameters (all of which are optional):

    - ``tm`` - The number of seconds since autonomous has started
    - ``state_tm`` - The number of seconds since this state has been active
      (note: it may not start at zero!)
    - ``initial_call`` - Set to True when the state is initially called,
      False otherwise. If the state is switched to multiple times, this
      will be set to True at the start of each state.

    To be consistent with the magicbot philosophy, in order for the
    state machine to execute its states you must call the :func:`engage`
    function upon each execution of the main robot control loop. If you do
    not call this function, then execution of the FSM will cease.

    .. note:: If you wish for the FSM to continue executing state functions
              regardless whether ``engage()`` is called, you must set the
              ``must_finish`` parameter in your state decorator to be True.

    When execution ceases (because ``engage()`` was not called), the
    :func:`done` function will be called and the FSM will be reset to the
    starting state. The state functions will not be called again unless
    ``engage`` is called.

    As a magicbot component, StateMachine contains an ``execute`` function that
    will be called on each control loop. All state execution occurs from
    within that function call. If you call other components from a
    StateMachine, you should ensure that your StateMachine is declared
    *before* the other components in your Robot class.

    .. warning:: As StateMachine already contains an execute function,
                 there is no need to define your own ``execute`` function for
                 a state machine component -- if you override ``execute``,
                 then the state machine may not work correctly. Instead,
                 use the :func:`@default_state <.default_state>` decorator.

    Here's a very simple example of how you might implement a shooter
    automation component that moves a ball into a shooter when the
    shooter is ready::

        class ShooterAutomation(magicbot.StateMachine):

            # Some other component
            shooter: Shooter
            ball_pusher: BallPusher

            def fire(self):
                """This is called from the main loop."""
                self.engage()

            @state(first=True)
            def begin_firing(self):
                """
                This function will only be called IFF fire is called and
                the FSM isn't currently in the 'firing' state. If fire
                was not called, this function will not execute.
                """
                self.shooter.enable()
                if self.shooter.ready():
                    self.next_state('firing')

            @timed_state(duration=1.0, must_finish=True)
            def firing(self):
                """
                Because must_finish=True, once the FSM has reached this state,
                this state will continue executing even if engage isn't called.
                """
                self.shooter.enable()
                self.ball_pusher.push()

            #
            # Note that there is no execute function defined as part of
            # this component
            #

        ...

        class MyRobot(magicbot.MagicRobot):
            shooter_automation: ShooterAutomation

            shooter: Shooter
            ball_pusher: BallPusher

            def teleopPeriodic(self):

                if self.joystick.getTrigger():
                    self.shooter_automation.fire()

    This object has a lot of really useful NetworkTables integration as well:

    - tunables are created in /components/NAME/state
      - state durations can be tuned here
      - The 'current state' is output as it happens
      - Descriptions and names of the states are here (for dashboard use)


    .. warning:: This object is not intended to be threadsafe and should not
                 be accessed from multiple threads
    '''

    VERBOSE_LOGGING = False

    #: A Python logging object automatically injected by magicbot.
    #: It can be used to send messages to the log, instead of using print statements.
    logger: logging.Logger

    #: NT variable that indicates which state will be executed next (though,
    #: does not guarantee that it will be executed). Will return an empty
    #: string if the state machine is not currently engaged.
    current_state = tunable("", subtable="state")

    state_names: ClassVar[tunable[Sequence[str]]]
    state_descriptions: ClassVar[tunable[Sequence[str]]]

    def __new__(cls) -> "StateMachine":
        # choose to use __new__ instead of __init__
        o = super().__new__(cls)
        o._build_states()
        return o

        # TODO: when this gets invoked, tunables need to be setup on
        # the object first

    def _build_states(self) -> None:
        has_first = False

        # problem: the user interface won't know which entries are the
        #          current variables being used by the robot. So, we setup
        #          an array with the names, and the dashboard uses that
        #          to determine the ordering too

        nt_names = []
        nt_desc = []

        states = {}
        cls = type(self)

        default_state = None

        # for each state function:
        for name, state in _get_class_members(cls).items():
            if not isinstance(state, _State):
                continue

            # is this the first state to execute?
            if state.first:
                if has_first:
                    raise MultipleFirstStatesError(
                        "Multiple states were specified as the first state!"
                    )

                self.__first = name
                has_first = True

            state_data = _StateData(state)
            states[name] = state_data
            nt_names.append(name)
            nt_desc.append(state.description or "")

            if state.is_default:
                if default_state is not None:
                    raise MultipleDefaultStatesError(
                        "Multiple default states are not allowed"
                    )
                default_state = state_data

        if not has_first:
            raise NoFirstStateError(
                "Starting state not defined! Use first=True on a state decorator"
            )

        # NOTE: this depends on tunables being bound after this function is called
        cls.state_names = tunable(nt_names, subtable="state")
        cls.state_descriptions = tunable(nt_desc, subtable="state")

        # Indicates that an external party wishes the state machine to execute
        self.__should_engage = False

        # Indicates that the state machine is currently executing
        self.__engaged = False

        # A dictionary of states
        self.__states = states

        # The currently executing state, or None if not executing
        self.__state: Optional[_StateData] = None

        # The default state
        self.__default_state = default_state

        # Variable to store time in
        self.__start = 0

    @property
    def is_executing(self) -> bool:
        """:returns: True if the state machine is executing states"""
        # return self.__state is not None
        return self.__engaged

    def on_enable(self) -> None:
        """
        magicbot component API: called when autonomous/teleop is enabled
        """
        pass

    def on_disable(self) -> None:
        """
        magicbot component API: called when autonomous/teleop is disabled
        """
        self.done()

    def engage(
        self,
        initial_state: Optional[StateRef] = None,
        force: bool = False,
    ) -> None:
        """
        This signals that you want the state machine to execute its
        states.

        :param initial_state: If specified and execution is not currently
                              occurring, start in this state instead of
                              in the 'first' state
        :param force:         If True, will transition even if the state
                              machine is currently active.
        """
        self.__should_engage = True

        if force or self.__state is None or self.__state is self.__default_state:
            if initial_state:
                self.next_state(initial_state)
            else:
                self.next_state(self.__first)

    def next_state(self, state: StateRef) -> None:
        """Call this function to transition to the next state

        :param state: Name of the state to transition to

        .. note:: This should only be called from one of the state functions
        """
        if isinstance(state, _State):
            state = state.name

        state_data = self.__states[state]
        state_data.ran = False
        self.current_state = state

        self.__state = state_data

    def next_state_now(self, state: StateRef) -> None:
        """Call this function to transition to the next state, and call the next
        state function immediately. Prefer to use :meth:`next_state` instead.

        :param state: Name of the state to transition to

        .. note:: This should only be called from one of the state functions
        """
        self.next_state(state)
        # TODO: may want to do this differently?
        self.execute()

    def done(self) -> None:
        """Call this function to end execution of the state machine.

        This function will always be called when a state machine ends. Even if
        the engage function is called repeatedly, done() will be called.

        .. note:: If you wish to do something each time execution ceases,
                  override this function (but be sure to call
                  ``super().done()``!)
        """
        if self.VERBOSE_LOGGING and self.__state is not None:
            self.logger.info("Stopped state machine execution")

        self.__state = None
        self.__engaged = False
        self.current_state = ""

    def execute(self) -> None:
        """
        magicbot component API: This is called on each iteration of the
        control loop. Most of the time, you will not want to override
        this function. If you find you want to, you may want to use the
        @default_state mechanism instead.
        """

        now = getTime()

        if not self.__engaged:
            if self.__should_engage:
                self.__start = now
                self.__engaged = True
            elif self.__default_state is None:
                return

        # tm is the number of seconds that the state machine has been executing
        tm = now - self.__start
        state = self.__state
        done_called = False

        # we adjust this so that if we have states chained together,
        # then the total time it runs is the amount of time of the
        # states. Otherwise, the time drifts.
        new_state_start = tm

        # determine if the time has passed to execute the next state
        # -> intentionally comes first
        if state is not None and state.ran and state.expires < tm:
            new_state_start = state.expires

            if state.next_state is None:
                # If the state expires and it's the last state, if the machine
                # is still engaged then it should cycle back to the beginning
                # ... but we should call done() first
                done_called = True
                self.done()

                if self.__should_engage:
                    self.next_state(self.__first)
                    state = self.__state
                else:
                    state = None
            else:
                self.next_state(state.next_state)
                state = self.__state

        # deactivate the current state unless engage was called or
        # must_finish was set
        if not (self.__should_engage or state is not None and state.must_finish):
            state = None

        # if there is no state to execute and there is a default
        # state, do the default state
        if state is None and self.__default_state is not None:
            state = self.__default_state
            if self.__state != state:
                state.ran = False
                self.__state = state

        if state is not None:
            # is this the first time this was executed?
            initial_call = not state.ran
            if initial_call:
                state.ran = True
                state.start_time = new_state_start
                state.expires = new_state_start + getattr(
                    self, state.duration_attr, 0xFFFFFFFF
                )

                if self.VERBOSE_LOGGING:
                    self.logger.info("%.3fs: Entering state: %s", tm, state.name)

            # execute the state function, passing it the arguments
            state.run(self, tm, tm - state.start_time, initial_call)
        elif not done_called:
            # or clear the state
            self.done()

        # Reset this each time
        self.__should_engage = False


class AutonomousStateMachine(StateMachine):
    """
    This is a specialized version of the StateMachine that is designed
    to be used as an autonomous mode. There are a few key differences:

    - The :func:`.engage` function is always called, so the state machine
      will always run to completion unless done() is called
    - VERBOSE_LOGGING is set to True, so a log message will be printed out upon
      each state transition

    """

    VERBOSE_LOGGING = True

    def on_enable(self) -> None:
        super().on_enable()
        self.__engaged = True

    def on_iteration(self, tm: float) -> None:
        # TODO, remove the on_iteration function in 2017?

        # Only engage the state machine until its execution finishes, otherwise
        # it will just keep repeating
        #
        # This is because if you keep calling engage(), the state machine will
        # loop. I'm tempted to change that, but I think it would lead to unexpected
        # side effects. Will have to contemplate this...

        if self.__engaged:
            self.engage()
            self.execute()
            self.__engaged = self.is_executing

    def done(self) -> None:
        super().done()
        self._StateMachine__should_engage = False
        self.__engaged = False



import wpilib
import logging
from typing import List, Tuple

logger = logging.getLogger("simple_watchdog")

__all__ = ["SimpleWatchdog"]


class SimpleWatchdog:
    """A class that's a wrapper around a watchdog timer.

    When the timer expires, a message is printed to the console and an optional user-provided
    callback is invoked.

    The watchdog is initialized disabled, so the user needs to call enable() before use.

    .. note:: This is a simpler replacement for the :class:`wpilib.Watchdog`,
              and should function mostly the same (except that this watchdog will
              not detect infinite loops).

    .. warning:: This watchdog is not threadsafe

    """

    # Used for timeout print rate-limiting
    kMinPrintPeriod = 1000000  # us

    def __init__(self, timeout: float):
        """Watchdog constructor.

        :param timeout: The watchdog's timeout in seconds with microsecond resolution.
        """
        self._get_time = wpilib.RobotController.getFPGATime

        self._startTime = 0  # us
        self._timeout = int(timeout * 1e6)  # us
        self._expirationTime = 0  # us
        self._lastTimeoutPrintTime = 0  # us
        self._lastEpochsPrintTime = 0  # us
        self._epochs: List[Tuple[str, int]] = []

    def getTime(self) -> float:
        """Returns the time in seconds since the watchdog was last fed."""
        return (self._get_time() - self._startTime) / 1e6

    def setTimeout(self, timeout: float) -> None:
        """Sets the watchdog's timeout.

        :param timeout: The watchdog's timeout in seconds with microsecond
                        resolution.
        """
        self._epochs.clear()
        timeout = int(timeout * 1e6)  # us
        self._timeout = timeout
        self._startTime = self._get_time()
        self._expirationTime = self._startTime + timeout

    def getTimeout(self) -> float:
        """Returns the watchdog's timeout in seconds."""
        return self._timeout / 1e6

    def isExpired(self) -> bool:
        """Returns true if the watchdog timer has expired."""
        return self._get_time() > self._expirationTime

    def addEpoch(self, epochName: str) -> None:
        """
        Adds time since last epoch to the list printed by printEpochs().

        Epochs are a way to partition the time elapsed so that when
        overruns occur, one can determine which parts of an operation
        consumed the most time.

        :param epochName: The name to associate with the epoch.
        """
        self._epochs.append((epochName, self._get_time()))

    def printIfExpired(self) -> None:
        """Prints list of epochs added so far and their times."""
        now = self._get_time()
        if (
            now > self._expirationTime
            and now - self._lastEpochsPrintTime > self.kMinPrintPeriod
        ):
            self._lastEpochsPrintTime = now
            prev = self._startTime
            logger.warning("Watchdog not fed after %.6fs", (now - prev) / 1e6)
            epoch_logs = []
            for key, value in self._epochs:
                time = (value - prev) / 1e6
                epoch_logs.append(f"\t{key}: {time:.6f}")
                prev = value
            logger.info("Epochs:\n%s", "\n".join(epoch_logs))

    def reset(self) -> None:
        """Resets the watchdog timer.

        This also enables the timer if it was previously disabled.
        """
        self.enable()

    def enable(self) -> None:
        """Enables the watchdog timer."""
        self._epochs.clear()
        self._startTime = self._get_time()
        self._expirationTime = self._startTime + self._timeout

    def disable(self) -> None:
        """Disables the watchdog timer."""
        # .. this doesn't do anything



def test_digital_sensor(wpimock):
    wpimock.Counter().getPeriod.return_value = 1 * 0.000147

    from robotpy_ext.common_drivers import xl_max_sonar_ez

    digital = xl_max_sonar_ez.MaxSonarEZPulseWidth(1)
    assert digital.get() == 1


def test_analog_sensor(wpimock):
    from robotpy_ext.common_drivers import xl_max_sonar_ez

    analog = xl_max_sonar_ez.MaxSonarEZAnalog(1)
    analog.analog.getVoltage.return_value = 1 * 0.0049 * 2.54

    assert analog.get() == 1



import math
import wpilib

from wpilib.simulation import AnalogInputSim

from .distance_sensors import SharpIR2Y0A02, SharpIR2Y0A21, SharpIR2Y0A41


class SharpIR2Y0A02Sim:
    """
    An easy to use simulation interface for a Sharp GP2Y0A02YK0F
    """

    def __init__(self, sensor: SharpIR2Y0A02) -> None:
        assert isinstance(sensor, SharpIR2Y0A02)
        self._sim = AnalogInputSim(sensor.distance)
        self._distance = 0

    def getDistance(self) -> float:
        """Get set distance (not distance sensor sees) in centimeters"""
        return self._distance

    def setDistance(self, d) -> None:
        """Set distance in centimeters"""
        self._distance = d
        d = max(min(d, 145.0), 22.5)
        v = math.pow(d / 62.28, 1 / -1.092)
        self._sim.setVoltage(v)


class SharpIR2Y0A21Sim:
    """
    An easy to use simulation interface for a Sharp GP2Y0A21YK0F
    """

    def __init__(self, sensor: SharpIR2Y0A21) -> None:
        assert isinstance(sensor, SharpIR2Y0A21)
        self._sim = AnalogInputSim(sensor.distance)
        self._distance = 0

    def getDistance(self) -> float:
        """Get set distance (not distance sensor sees) in centimeters"""
        return self._distance

    def setDistance(self, d) -> None:
        """Set distance in centimeters"""
        self._distance = d
        d = max(min(d, 80.0), 10.0)
        v = math.pow(d / 26.449, 1 / -1.226)
        self._sim.setVoltage(v)


class SharpIR2Y0A41Sim:
    """
    An easy to use simulation interface for a Sharp GP2Y0A41SK0F
    """

    def __init__(self, sensor: SharpIR2Y0A41) -> None:
        assert isinstance(sensor, SharpIR2Y0A41)
        self._sim = AnalogInputSim(sensor.distance)
        self._distance = 0

    def getDistance(self) -> float:
        """Get set distance (not distance sensor sees) in centimeters"""
        return self._distance

    def setDistance(self, d) -> None:
        """Set distance in centimeters"""
        self._distance = d
        d = max(min(d, 35.0), 4.5)
        v = math.pow(d / 12.84, 1 / -0.9824)
        self._sim.setVoltage(v)



import logging
from typing import Any, Dict, Optional

logger = logging.getLogger(__name__)


class MagicInjectError(ValueError):
    pass


def get_injection_requests(
    type_hints: Dict[str, type], cname: str, component: Optional[Any] = None
) -> Dict[str, type]:
    """
    Given a dict of type hints, filter it to the requested injection types.

    :param type_hints: The type hints to inspect.
    :param cname: The component name.
    :param component: The component if it has been instantiated.
    """
    requests = {}

    for n, inject_type in type_hints.items():
        # If the variable is private ignore it
        if n.startswith("_"):
            if component is None:
                message = f"Cannot inject into component {cname} __init__ param {n}"
                raise MagicInjectError(message)
            continue

        # If the variable has been set, skip it
        if component is not None and hasattr(component, n):
            continue

        # Check for generic types from the typing module
        origin = getattr(inject_type, "__origin__", None)
        if origin is not None:
            inject_type = origin

        # If the type is not actually a type, give a meaningful error
        if not isinstance(inject_type, type):
            message = (
                f"Component {cname} has a non-type annotation {n}: {inject_type!r}"
            )
            if component is not None:
                message += "\nLone non-injection variable annotations are disallowed. Did you mean to assign a static variable?"
            raise TypeError(message)

        requests[n] = inject_type

    return requests


def find_injections(
    requests: Dict[str, type], injectables: Dict[str, Any], cname: str
) -> Dict[str, Any]:
    """
    Get a dict of the variables to inject into a given component.

    :param requests: The mapping of requested variables to types,
                     as returned by :func:`get_injection_requests`.
    :param injectables: The available variables to inject.
    :param cname: The name of the component.
    """
    to_inject = {}

    for n, inject_type in requests.items():
        injectable = injectables.get(n)
        if injectable is None:
            # Try prefixing the component name
            injectable = injectables.get(f"{cname}_{n}")

        # Raise error if injectable syntax used but no injectable was found.
        if injectable is None:
            raise MagicInjectError(
                "Component %s has variable %s (type %s), which is absent from robot"
                % (cname, n, inject_type)
            )

        # Raise error if injectable declared with type different than the initial type
        if not isinstance(injectable, inject_type):
            raise MagicInjectError(
                "Component %s variable %s does not match type in robot! (Got %s, expected %s)"
                % (cname, n, type(injectable), inject_type)
            )

        to_inject[n] = injectable
        logger.debug("-> %s.%s = %s", cname, n, injectable)

    return to_inject



import typing

import pytest

from magicbot.inject import get_injection_requests


def test_ctor_invalid_type_hint_message():
    """
    class Component:
        def __init__(self, foo: 1): ...
    """
    type_hints = {
        "foo": typing.cast(type, 1),
    }

    with pytest.raises(TypeError) as exc_info:
        get_injection_requests(type_hints, "bar")

    assert exc_info.value.args[0] == "Component bar has a non-type annotation foo: 1"



import functools
import inspect
import logging
from typing import Callable, Optional

import ntcore
import wpilib

logger = logging.getLogger("autonomous")


# use this to track ordering of functions, so that we can display them
# properly in the tuning widget on the dashboard
__global_cnt_serial = [0]


def _get_state_serial():
    __global_cnt_serial[0] = __global_cnt_serial[0] + 1
    return __global_cnt_serial[0]


class _State:
    def __init__(self, f: Callable, first: bool):
        # inspect the args, provide a correct call implementation
        allowed_args = "self", "tm", "state_tm", "initial_call"
        sig = inspect.signature(f)
        name = f.__name__

        args = []
        invalid_args = []
        for i, arg in enumerate(sig.parameters.values()):
            if i == 0 and arg.name != "self":
                raise ValueError(f"First argument to {name} must be 'self'")
            if arg.kind is arg.VAR_POSITIONAL:
                raise ValueError(f"Cannot use *args in signature for function {name}")
            if arg.kind is arg.VAR_KEYWORD:
                raise ValueError(
                    f"Cannot use **kwargs in signature for function {name}"
                )
            if arg.kind is arg.KEYWORD_ONLY:
                raise ValueError(
                    "Cannot use keyword-only parameters for function %s" % name
                )
            if arg.name in allowed_args:
                args.append(arg.name)
            else:
                invalid_args.append(arg.name)

        if invalid_args:
            raise ValueError(
                "Invalid parameter names in %s: %s" % (name, ",".join(invalid_args))
            )

        functools.update_wrapper(self, f)

        # store state variables here
        self._func = f
        self.name = name
        self.description = f.__doc__
        self.ran = False
        self.first = first
        self.expires = 0xFFFFFFFF
        self.serial = _get_state_serial()

        varlist = {"f": f}
        args_code = ",".join(args)
        wrapper_code = f"lambda self, tm, state_tm, initial_call: f({args_code})"
        self.run = eval(wrapper_code, varlist, varlist)

    def __call__(self, *args, **kwargs):
        self._func(*args, **kwargs)

    def __set_name__(self, owner: type, name: str) -> None:
        if not issubclass(owner, StatefulAutonomous):
            raise TypeError(
                f"StatefulAutonomous state {name} defined in non-StatefulAutonomous"
            )


#
# Decorators:
#
#   state
#   timed_state
#


def timed_state(
    f: Optional[Callable] = None,
    duration: float = None,
    next_state: str = None,
    first: bool = False,
):
    """
    If this decorator is applied to a function in an object that inherits
    from :class:`.StatefulAutonomous`, it indicates that the function
    is a state that will run for a set amount of time unless interrupted

    The decorated function can have the following arguments in any order:

    - ``tm`` - The number of seconds since autonomous has started
    - ``state_tm`` - The number of seconds since this state has been active
      (note: it may not start at zero!)
    - ``initial_call`` - Set to True when the state is initially called,
      False otherwise. If the state is switched to multiple times, this
      will be set to True at the start of each state.

    :param duration: The length of time to run the state before progressing
                     to the next state
    :param next_state: The name of the next state. If not specified, then
                       this will be the last state executed if time expires
    :param first: If True, this state will be ran first
    """

    if f is None:
        return functools.partial(
            timed_state, duration=duration, next_state=next_state, first=first
        )

    if duration is None:
        raise ValueError("timed_state functions must specify a duration")

    wrapper = _State(f, first)

    wrapper.next_state = next_state
    wrapper.duration = duration

    return wrapper


def state(f: Optional[Callable] = None, first: bool = False):
    """
    If this decorator is applied to a function in an object that inherits
    from :class:`.StatefulAutonomous`, it indicates that the function
    is a state. The state will continue to be executed until the
    ``next_state`` function is executed.

    The decorated function can have the following arguments in any order:

    - ``tm`` - The number of seconds since autonomous has started
    - ``state_tm`` - The number of seconds since this state has been active
      (note: it may not start at zero!)
    - ``initial_call`` - Set to True when the state is initially called,
      False otherwise. If the state is switched to multiple times, this
      will be set to True at the start of each state.

    :param first: If True, this state will be ran first
    """

    if f is None:
        return functools.partial(state, first=first)

    return _State(f, first)


class StatefulAutonomous:
    """
    This object is designed to be used to implement autonomous modes that
    can be used with the :class:`.AutonomousModeSelector` object to select
    an appropriate autonomous mode. However, you don't have to.

    This object is designed to meet the following goals:

    - Supports simple built-in tuning of autonomous mode parameters via
      SmartDashboard
    - Easy to create autonomous modes that support state machine or
      time-based operation
    - Autonomous modes that are easy to read and understand

    You use this by defining a class that inherits from ``StatefulAutonomous``.
    To define each state, you use the :func:`timed_state` decorator on a
    function. When each state is run, the decorated function will be
    called. Decorated functions can receive the following parameters:

    - ``tm`` - The number of seconds since autonomous has started
    - ``state_tm`` - The number of seconds since this state has been active
      (note: it may not start at zero!)
    - ``initial_call`` - Set to True when the state is initially called,
      False otherwise. If the state is switched to multiple times, this
      will be set to True at the start of each state.

    An example autonomous mode that drives the robot forward for 5 seconds
    might look something like this::

        from robotpy_ext.autonomous import StatefulAutonomous

        class DriveForward(StatefulAutonomous):

            MODE_NAME = 'Drive Forward'

            def initialize(self):
                pass

            @timed_state(duration=0.5, next_state='drive_forward', first=True)
            def drive_wait(self):
                pass

            @timed_state(duration=5)
            def drive_forward(self):
                self.drive.move(0, 1, 0)

    Note that in this example, it is assumed that the DriveForward object
    is initialized with a dictionary with a value 'drive' that contains
    an object that has a move function::

        components = {'drive': SomeObject() }
        mode = DriveForward(components)

    If you use this object with :class:`.AutonomousModeSelector`, make sure
    to initialize it with the dictionary, and it will be passed to this
    autonomous mode object when initialized.

    .. seealso:: Check out the samples in our github repository that show
                 some basic usage of ``AutonomousModeSelector``.
    """

    __built = False
    __done = False

    def __init__(self, components=None):
        """
        :param components: A dictionary of values that will be assigned
                           as attributes to this object, using the key
                           names in the dictionary
        :type  components: dict
        """

        if not hasattr(self, "MODE_NAME"):
            raise ValueError("Must define MODE_NAME class variable")

        if components:
            for k, v in components.items():
                setattr(self, k, v)

        NetworkTables = ntcore.NetworkTableInstance.getDefault()
        self.__table = NetworkTables.getTable("SmartDashboard")
        self.__sd_args = []

        self.__build_states()
        self.__tunables = []

        if hasattr(self, "initialize"):
            self.initialize()

    def register_sd_var(self, name, default, add_prefix=True, vmin=-1, vmax=1):
        """
        Register a variable that is tunable via NetworkTables/SmartDashboard

        When this autonomous mode is enabled, all of the SmartDashboard
        settings will be read and stored as attributes of this object. For
        example, to register a variable 'foo' with a default value of 1::

            self.register_sd_var('foo', 1)

        This value will show up on NetworkTables as the key ``MODE_NAME\\foo``
        if add_prefix is specified, otherwise as ``foo``.

        :param name:     Name of variable to display to user, cannot have a
                         space in it.
        :param default:  Default value of variable
        :param add_prefix: Prefix this setting with the mode name
        :type  add_prefix: bool
        :param vmin:     For tuning: minimum value of this variable
        :param vmax:     For tuning: maximum value of this variable
        """

        is_number = self.__register_sd_var_internal(name, default, add_prefix, True)

        if not add_prefix:
            return

        # communicate the min/max value for numbers to the dashboard
        if is_number:
            name = "%s|%0.3f|%0.3f" % (name, vmin, vmax)

        self.__tunables.append(name)
        self.__table.putStringArray(self.MODE_NAME + "_tunables", self.__tunables)

    def __register_sd_var_internal(self, name, default, add_prefix, readback):
        if " " in name:
            raise ValueError(
                "ERROR: Cannot use spaces in a tunable variable name (%s)" % name
            )

        is_number = False
        sd_name = name

        if add_prefix:
            sd_name = "%s\\%s" % (self.MODE_NAME, name)

        if isinstance(default, bool):
            self.__table.putBoolean(sd_name, default)
            args = (name, sd_name, self.__table.getBoolean, default)

        elif isinstance(default, int) or isinstance(default, float):
            self.__table.putNumber(sd_name, default)
            args = (name, sd_name, self.__table.getNumber, default)
            is_number = True

        elif isinstance(default, str):
            self.__table.putString(sd_name, default)
            args = (name, sd_name, self.__table.getString, default)

        else:
            raise ValueError("Invalid default value")

        if readback:
            self.__sd_args.append(args)
        return is_number

    def __build_states(self):
        has_first = False

        states = {}
        cls = type(self)

        # for each state function:
        for name in dir(cls):
            state = getattr(cls, name)
            if not isinstance(state, _State):
                continue

            # is this the first state to execute?
            if state.first:
                if has_first:
                    raise ValueError(
                        "Multiple states were specified as the first state!"
                    )

                self.__first = name
                has_first = True

            # problem: how do we expire old entries?
            # -> what if we just use json? more flexible, but then we can't tune it
            #    via SmartDashboard

            # make the time tunable
            if hasattr(state, "duration"):
                self.__register_sd_var_internal(
                    state.name + "_duration", state.duration, True, True
                )

            description = ""
            if state.description is not None:
                description = state.description

            states[state.serial] = (state.name, description)

        # problem: the user interface won't know which entries are the
        #          current variables being used by the robot. So, we setup
        #          an array with the names, and the dashboard uses that
        #          to determine the ordering too

        sorted_states = sorted(states.items())

        self.__table.putStringArray(
            self.MODE_NAME + "_durations", [name for _, (name, desc) in sorted_states]
        )

        self.__table.putStringArray(
            self.MODE_NAME + "_descriptions",
            [desc for _, (name, desc) in sorted_states],
        )

        if not has_first:
            raise ValueError(
                "Starting state not defined! Use first=True on a state decorator"
            )

        self.__built = True

    def _validate(self):
        # TODO: make sure the state machine can be executed
        # - run at robot time? Probably not. Run this as part of a unit test
        pass

    # how long does introspection take? do this in the constructor?

    # can do things like add all of the timed states, and project how long
    # it will take to execute it (don't forget about cycles!)

    def on_enable(self):
        """
        Called when autonomous mode is enabled, and initializes the
        state machine internals.

        If you override this function, be sure to call it from your
        customized ``on_enable`` function::

            super().on_enable()
        """

        if not self.__built:
            raise ValueError("super().__init__(components) was never called!")

        # print out the details of this autonomous mode, and any tunables

        self.battery_voltage = wpilib.DriverStation.getBatteryVoltage()
        logger.info("Battery voltage: %.02fv", self.battery_voltage)

        logger.info("Tunable values:")

        # read smart dashboard values, print them
        for name, sd_name, fn, default in self.__sd_args:
            val = fn(sd_name, default)
            setattr(self, name, val)
            logger.info("-> %25s: %s" % (name, val))

        # set the starting state
        self.next_state(self.__first)
        self.__done = False

    def on_disable(self):
        """Called when the autonomous mode is disabled"""
        pass

    def done(self):
        """Call this function to indicate that no more states should be called"""
        self.next_state(None)

    def next_state(self, name):
        """Call this function to transition to the next state

        :param name: Name of the state to transition to
        """
        if name is not None:
            self.__state = getattr(self.__class__, name)
        else:
            self.__state = None

        if self.__state is None:
            return

        self.__state.ran = False

    def on_iteration(self, tm):
        """This function is called by the autonomous mode switcher, should
        not be called by enduser code. It is called once per control
        loop iteration."""

        # if you get an error here, then you probably overrode on_enable,
        # but didn't call super().on_enable(). Don't do that.
        try:
            state = self.__state
        except AttributeError:
            raise ValueError("super().on_enable was never called!")

        # we adjust this so that if we have states chained together,
        # then the total time it runs is the amount of time of the
        # states. Otherwise, the time drifts.
        new_state_start = tm

        # determine if the time has passed to execute the next state
        if state is not None and state.expires < tm:
            self.next_state(state.next_state)
            new_state_start = state.expires
            state = self.__state

        if state is None:
            if not self.__done:
                logger.info("%.3fs: Done with autonomous mode", tm)
                self.__done = True
            return

        # is this the first time this was executed?
        initial_call = not state.ran
        if initial_call:
            state.ran = True
            state.start_time = new_state_start
            state.expires = state.start_time + getattr(
                self, state.name + "_duration", 0xFFFFFFFF
            )

            logger.info("%.3fs: Entering state: %s", tm, state.name)

        # execute the state function, passing it the arguments
        state.run(self, tm, tm - state.start_time, initial_call)



from robotpy_ext.common_drivers.distance_sensors import (
    SharpIR2Y0A02,
    SharpIR2Y0A21,
    SharpIR2Y0A41,
)

from robotpy_ext.common_drivers.distance_sensors_sim import (
    SharpIR2Y0A02Sim,
    SharpIR2Y0A21Sim,
    SharpIR2Y0A41Sim,
)

import pytest


def test_2Y0A02(hal):
    sensor = SharpIR2Y0A02(0)
    sim = SharpIR2Y0A02Sim(sensor)

    # 22 to 145cm

    # min
    sim.setDistance(10)
    assert sensor.getDistance() == pytest.approx(22.5)

    # max
    sim.setDistance(200)
    assert sensor.getDistance() == pytest.approx(145)

    # middle
    sim.setDistance(50)
    assert sensor.getDistance() == pytest.approx(50)

    sim.setDistance(100)
    assert sensor.getDistance() == pytest.approx(100)


def test_2Y0A021(hal):
    sensor = SharpIR2Y0A21(0)
    sim = SharpIR2Y0A21Sim(sensor)

    # 10 to 80cm

    # min
    sim.setDistance(5)
    assert sensor.getDistance() == pytest.approx(10)

    # max
    sim.setDistance(100)
    assert sensor.getDistance() == pytest.approx(80)

    # middle
    sim.setDistance(30)
    assert sensor.getDistance() == pytest.approx(30)

    sim.setDistance(60)
    assert sensor.getDistance() == pytest.approx(60)


def test_2Y0A041(hal):
    sensor = SharpIR2Y0A41(0)
    sim = SharpIR2Y0A41Sim(sensor)

    # 4.5 to 35 cm

    # min
    sim.setDistance(2)
    assert sensor.getDistance() == pytest.approx(4.5)

    # max
    sim.setDistance(50)
    assert sensor.getDistance() == pytest.approx(35)

    # middle
    sim.setDistance(10)
    assert sensor.getDistance() == pytest.approx(10)

    sim.setDistance(25)
    assert sensor.getDistance() == pytest.approx(25)
